using System;
using System.Text;

namespace Org.BouncyCastle.Asn1
{
    /**
     * Der T61String (also the teletex string)
     */
    public class DerT61String
        : DerStringBase
    {
        private readonly string str;

        /**
         * return a T61 string from the passed in object.
         *
         * @exception ArgumentException if the object cannot be converted.
         */
        public static DerT61String GetInstance(
            object obj)
        {
            if (obj == null || obj is DerT61String)
            {
                return (DerT61String)obj;
            }

            if (obj is Asn1OctetString)
            {
                return new DerT61String(((Asn1OctetString)obj).GetOctets());
            }

            if (obj is Asn1TaggedObject)
            {
                return GetInstance(((Asn1TaggedObject)obj).GetObject());
            }

            throw new ArgumentException("illegal object in GetInstance: " + obj.GetType().Name);
        }

        /**
         * return an T61 string from a tagged object.
         *
         * @param obj the tagged object holding the object we want
         * @param explicitly true if the object is meant to be explicitly
         *              tagged false otherwise.
         * @exception ArgumentException if the tagged object cannot
         *               be converted.
         */
        public static DerT61String GetInstance(
            Asn1TaggedObject	obj,
            bool				explicitly)
        {
            return GetInstance(obj.GetObject());
        }

        /**
         * basic constructor - with bytes.
         */
        public DerT61String(
            byte[] str)
			: this(Encoding.ASCII.GetString(str))
        {
        }

		/**
         * basic constructor - with string.
         */
        public DerT61String(
            string str)
        {
			if (str == null)
				throw new ArgumentNullException("str");

			this.str = str;
        }

		public override string GetString()
        {
            return str;
        }

        internal override void Encode(
            DerOutputStream derOut)
        {
            derOut.WriteEncoded(Asn1Tags.T61String, this.GetOctets());
        }

        public byte[] GetOctets()
        {
			return Encoding.ASCII.GetBytes(str);
        }

		protected override bool Asn1Equals(
			Asn1Object obj)
		{
			DerT61String other = obj as DerT61String;

			if (other == null)
				return false;

            return this.str.Equals(other.str);
        }
    }
}
